/**
 * 
 *  In game menu of other servers and their status.
 *
 *  To set the list of servers available, create a file serverlist.cfg in the
 *  amxx config directory. Each line should contain one server and look like:
 *
 *  Server name\IP or hostname:port password
 *
 *  Server name is whatever you want the server to be called in the menu. That
 *  is followed by a backslash (\). Then the hostname or IP of the server
 *  followed by a colon (:) and then the port. If needed the password is put
 *  after a space following the port. Warning, this will allow anyone that
 *  uses the menu to connect to the passworded server to learn the password.
 *  The port and password are optional. The port will default to 27015 and
 *  must be defined if a password is used.
 * 
 *  Lines starting with # or ; are considered a comment. 
 *
 *  1.0
 *    Some small refactoring to clean up the code a bit.
 *    Fixed a bug where it would not handle a challenge with a 0x00 byte in it.
 *    Unfortunately this uncovered a similiar bug in the sockets module. So...
 *    Added a little more graceful handling of packet errors.
 *
 *  0.6
 *    Update to new steam server query protocol.
 *
 *  0.5
 *    Changed the way packets are formed to be sent. Hopefully working around
 *    the linux compiler bug and also being 64bit compliant.
 *      Suggested by Bailopan.
 *
 *  0.4
 *    Fixed showing incorrect number of pages on the main server menu.
 *
 *  0.3:
 *    Changed includes to conform with standard practice
 *    Made the port optional in the server list
 *    Added ability to connect to password protected servers
 *
 *  0.2:
 *    Initial public release
 *
 *  Plugin developed by Janzert aka Brian Haskin
 *
 *  Copyright 2004, 2005 Brian Haskin
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  In addition, as a special exception, the author gives permission to
 *  link the code of this program with the Half-Life Game Engine ("HL
 *  Engine") and Modified Game Libraries ("MODs") developed by Valve,
 *  L.L.C ("Valve"). You must obey the GNU General Public License in all
 *  respects for all of the code used other than the HL Engine and MODs
 *  from Valve. If you modify this file, you may extend this exception
 *  to your version of the file, but you are not obligated to do so. If
 *  you do not wish to do so, delete this exception statement from your
 *  version.
 **/

#include <amxmodx>
#include <amxmisc>
#include <sockets>

#define CONFIG_FILE "serverlist.cfg"
#define MAX_SERVERS 20
#define MAX_SERVER_NAME 40
#define MAX_SERVER_HOSTNAME 40
#define MAX_SERVER_PASSWORD 20
#define MAX_MAP_NAME 25
#define MAX_TIME 7

new numServers = 0
new serverName[MAX_SERVERS][MAX_SERVER_NAME]
new serverHost[MAX_SERVERS][MAX_SERVER_HOSTNAME]
new serverPort[MAX_SERVERS]
new serverPassword[MAX_SERVERS][MAX_SERVER_PASSWORD]
new serverLastQuery[MAX_SERVERS]
new serverPlayers[MAX_SERVERS]
new serverMaxPlayers[MAX_SERVERS]
new serverMapName[MAX_SERVERS][MAX_MAP_NAME]
new serverTimeleft[MAX_SERVERS][MAX_TIME]

new menuPosition[33]
new pendingQuery[33]
new lastServer[33]

#define MIN_QUERY_INTERVAL 60
#define MAX_QUERIES 5
new curQuery = 0
new lastQuery = 0
new queryPlayer[MAX_QUERIES]
new queryServer[MAX_QUERIES]

new socket = 0
new recvattempts = 0

#define BUF_SIZE 5000
new packet_buffer[BUF_SIZE]
new query_in_progress = 0
new multipart = 0
new num_of_parts = 0
new partsreceived = 0

public plugin_init()
{
  register_plugin("Server menu", "1.0", "Janzert")
  register_clcmd("say /servers", "startMenu", ADMIN_ALL, "Display other server information")
  register_menucmd(register_menuid("\yServers menu:"), 1023, "menuAction")
  register_menucmd(register_menuid("\yServer info:"), 1023, "serverAction")
  read_config()

  return PLUGIN_CONTINUE
}

read_config()
{
  new cfilename[80], configsdir[80]
  
  get_configsdir(configsdir, 79)
  format(cfilename, 79, "%s/%s", configsdir, CONFIG_FILE)
  
  if (!file_exists(cfilename))
  {
    log_amx("Couldn't read server list file.")
    return 0
  }

  new lineindex, linetext[256], linelength, serverindex
  
  while((serverindex < MAX_SERVERS) && read_file(cfilename,lineindex++,linetext,255, linelength))
  {
    linelength -= trim(linetext)
    if (linelength == 0)
      continue

    if ((linetext[0] == ';') || (linetext[0] == '#'))
      continue

    new sname[MAX_SERVER_NAME], shost[MAX_SERVER_HOSTNAME]
    new texti = 0
    while ((texti<linelength) && (linetext[texti]!='\')) 
    {
      texti++
    }
    new namelen = copyc(sname, MAX_SERVER_NAME, linetext, '\')
    if (namelen==MAX_SERVER_NAME)
    {
      log_amx("Server name end not found in %d characters on line %d", MAX_SERVER_NAME-1, lineindex+1)
      continue
    } else if (namelen==linelength)
    {
      log_amx("Couldn't get server name on line %d.", lineindex+1)
      continue
    }
    trim(sname)
    
    new hostlen = copyc(shost, MAX_SERVER_HOSTNAME, linetext[namelen+1], ':')
    new hostend = namelen + hostlen + 1
    if (hostlen == MAX_SERVER_HOSTNAME)
    {
      log_amx("Hostname end not found in %d characters on line %d.", MAX_SERVER_HOSTNAME-1, lineindex+1)
      continue
    }
    trim(shost)

    copy(serverName[numServers], MAX_SERVER_NAME, sname)
    copy(serverHost[numServers], MAX_SERVER_HOSTNAME, shost)

    new sport = 27015
    if (hostend < linelength)
    {
      new portstr[7]
      new portend = hostend + copyc(portstr, 6, linetext[hostend+1], ' ') + 1
      sport = str_to_num(portstr)
      serverPort[numServers] = sport

      if (portend < linelength)
      {
        copy(serverPassword[numServers], MAX_SERVER_PASSWORD, linetext[portend+1])
        trim(serverPassword[numServers])
      }
    }

    numServers += 1
  }

  return 1
}

#define MAX_MENU_LEN 8

showMenu(id)
{
  new startPos = menuPosition[id] * MAX_MENU_LEN
  if (startPos >= numServers)
  {
    startPos = menuPosition[id] = 0
  }
  new endPos = clamp(startPos+MAX_MENU_LEN, 0, numServers)
  new maxPages = (numServers%MAX_MENU_LEN)?(numServers/MAX_MENU_LEN)+1:numServers/MAX_MENU_LEN

  new menuString[500]
  new mStrLen = format(menuString, 500, "\yServers menu: \r%d/%d^n\w", menuPosition[id]+1, maxPages)

  new serveri, item, menuKeys = 0
  for (serveri = startPos; serveri < endPos; serveri++)
  {
    menuKeys |= (1<<item++)
    mStrLen += format(menuString[mStrLen], (500-mStrLen), "%d. %s^n", item, serverName[serveri])
  }

  if (endPos != numServers)
  {
    menuKeys |= (1<<8)|(1<<9)
    mStrLen += format(menuString[mStrLen], (500-mStrLen), "^n9. More...^n0. %s", menuPosition[id]?"Back":"Exit")
  } else
  {
    menuKeys |= (1<<9)
    mStrLen += format(menuString[mStrLen], (500-mStrLen), "^n0. %s", menuPosition[id]?"Back":"Exit")
  }

  show_menu(id, menuKeys, menuString, -1)
  return PLUGIN_CONTINUE
}

public menuAction(id, key)
{
  switch(key)
  {
    case 8: 
    {
      menuPosition[id]++
      showMenu(id)
    }
    case 9:
      if (menuPosition[id])
      {
        menuPosition[id]--
        showMenu(id)
      }
    default:
      startServer(id, key+(menuPosition[id]*MAX_MENU_LEN))
  }

  return PLUGIN_HANDLED
}

public startMenu(id)
{
  if (pendingQuery[id])
  {
    client_print(id, print_chat, "Sorry you already have a query in progress. %d, %d", curQuery, lastQuery)
    return PLUGIN_CONTINUE
  }

  if (!numServers)
  {
    client_print(id, print_chat, "Sorry no servers configured.")
    return PLUGIN_CONTINUE
  }

  menuPosition[id] = 0
  showMenu(id)
  
  return PLUGIN_CONTINUE
}

startServer(id, serverindex)
{
  if (serverLastQuery[serverindex]+MIN_QUERY_INTERVAL > get_systime())
  {
    showServer(id, serverindex)
    return PLUGIN_CONTINUE
  }
  
  if ((lastQuery+1)%MAX_QUERIES != curQuery)
  {
    pendingQuery[id] = 1
    queryPlayer[lastQuery] = id
    queryServer[lastQuery] = serverindex
    lastQuery = (lastQuery+1)%MAX_QUERIES
    
    if (!query_in_progress)
      startQuery()
  } else
  {
    log_amx("Max queries in progress at once encountered.")
  }

  return PLUGIN_CONTINUE
}

startQuery()
{
  if (query_in_progress)
  {
    return PLUGIN_HANDLED
  }

  if (socket)
  {
    socket_close(socket)
  }

  new serveri = queryServer[curQuery]
  new error
  socket = socket_open(serverHost[serveri], serverPort[serveri], SOCKET_UDP, error)
  if (error)
  {
    switch (error)
    {
      case 1:
        log_amx("Error creating socket.")
      case 2:
        log_amx("Could not find server %s.", serverHost[serveri])
      case 3:
        // Should never happen with UDP.
        log_amx("Could not connect to server.")
    }
    socket = 0
    client_print(queryPlayer[curQuery], print_chat, "Sorry the query could not be completed.")
    pendingQuery[queryPlayer[curQuery]] = 0
    curQuery = (curQuery+1)%MAX_QUERIES
    if (curQuery!=lastQuery)
    {
      set_task(0.1, "startQuery")
    }
    return PLUGIN_HANDLED
  }

  query_in_progress = 1
  new packetstr[25]
  format(packetstr, 25, "%c%c%c%cTSource Engine Query", -1,-1,-1,-1)
  socket_send2(socket, packetstr, 25)
  set_task(0.2, "finish_query")
  
  return PLUGIN_HANDLED
}

public finish_query()
{
  if (!socket)
  {
    log_amx("finish_query called with null socket")
    client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
    query_in_progress = 0
    pendingQuery[queryPlayer[curQuery]] = 0
    curQuery = (curQuery+1)%MAX_QUERIES
    if (curQuery != lastQuery)
      set_task(0.1, "startQuery")
    return PLUGIN_CONTINUE
  }
  
  if (!socket_change(socket, 1))
  {
    recvattempts += 1
    if (recvattempts > 5)
    {
      log_amx("No response to info query.")
      client_print(queryPlayer[curQuery], print_chat, "No response received from server.")
      abort_query()
      return PLUGIN_CONTINUE
    } else
    {
      set_task(0.2, "finish_query")
      return PLUGIN_CONTINUE
    }
  }
  recvattempts = 0

  new packet[1450]
  socket_recv(socket, packet, 1450)
 // if (!equal(packet, {-1,-1,-1,-1,'m'}, 5))
 // {
 //   packet[6] = 0
 //   log_amx("Invalid packet received to info query. %s", packet)
 //  client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
 //   abort_query()
  //  return PLUGIN_CONTINUE
  //}

  new serveri = queryServer[curQuery]
  new strnum
  new packeti = 5
  for (strnum = 0; strnum < 5; strnum++)
  {
    if (strnum == 2)
    {
      copy(serverMapName[serveri], MAX_MAP_NAME, packet[packeti])
    }
    while((packeti < 1450) && (packet[packeti] != 0))
    {
      packeti += 1
    }
    packeti += 1
  }
  if (packeti >= 1450)
  {
    log_amx("Packet larger than buffer.")
    client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
    abort_query()
    return PLUGIN_CONTINUE
  }

  serverPlayers[serveri] = packet[packeti]
  serverMaxPlayers[serveri] = packet[packeti+1]

  new packetstr[5]
  format(packetstr, 5, "%c%c%c%cW", -1,-1,-1,-1)
  socket_send(socket, packetstr, 5)
  recvattempts = 0
  set_task(0.2, "query_challenge")
  return PLUGIN_CONTINUE
}

public query_challenge()
{
  if (!socket)
  {
    log_amx("query_challenge called with null socket")
    client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
    abort_query()
    return PLUGIN_CONTINUE
  }

  if (!socket_change(socket, 1))
  {
    recvattempts += 1
    if (recvattempts > 5)
    {
      log_amx("No response to challenge request.")
      client_print(queryPlayer[curQuery], print_chat, "No response received from server.")
      abort_query()
      return PLUGIN_CONTINUE
    } else
    {
      set_task(0.2, "query_time")
      return PLUGIN_CONTINUE
    }
  }
  recvattempts = 0

  new packet[1450]
  socket_recv(socket, packet, 1450)
  if (!equal(packet, {-1,-1,-1,-1,'A'}, 5))
  {
    packet[6] = 0
    log_amx("Invalid packet received for challenge response. %s", packet)
    client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
    abort_query()
    return PLUGIN_CONTINUE
  }

  new packetstr[10]
  format(packetstr, 5, "%c%c%c%cV", -1,-1,-1,-1)
  new i
  for (i=0; i<4; i++)
  {
    packetstr[i+5] = packet[i+5]
  }
  socket_send2(socket, packetstr, 9)
  recvattempts = 0
  set_task(0.2, "query_time")
  return PLUGIN_CONTINUE
}

public query_time()
{
  if (!socket)
  {
    log_amx("query_time called with null socket")
    client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
    abort_query()
    return PLUGIN_CONTINUE
  }
  
  if (!socket_change(socket, 1))
  {
    recvattempts += 1
    if (recvattempts > 5)
    {
      log_amx("No response to rules query for server %s:%d.", serverHost[queryServer[curQuery]], 
      serverPort[queryServer[curQuery]])
      serverTimeleft[queryServer[curQuery]][0] = 0
      show_query()
      return PLUGIN_CONTINUE
    } else
    {
      set_task(0.2, "query_time")
      return PLUGIN_CONTINUE
    }
  }
  recvattempts = 0

  new packet[1450]
  new len = socket_recv(socket, packet, 1450)
  if (equal(packet, {-2, -1, -1, -1}, 4))
  {
    new totalpackets = (packet[8] & 0x0f)
    new partindex = (packet[8] & 0xf0) >> 4
    if (!multipart)
    {
      multipart = 1
      num_of_parts = totalpackets
      partsreceived = 0
    } 
	else if (multipart && (totalpackets != num_of_parts))
    {
      log_amx("Total packets changed during multipacket receive.")
      client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
      socket_close(socket)
      abort_query()
      return PLUGIN_CONTINUE
    }
    
    new bufindex = partindex * 1391
    new packeti = 9
    while ((bufindex < BUF_SIZE) && (packeti < len))
    {
      packet_buffer[bufindex++] = packet[packeti++]
    }

    partsreceived += 1
    if (partsreceived == num_of_parts)
    {
      show_query()
      return PLUGIN_CONTINUE
    }
      
    set_task(0.1, "query_time")
    return PLUGIN_CONTINUE
  }

  new packeti
  for (packeti = 0; packeti < len; packeti++)
  {
    packet_buffer[packeti] = packet[packeti]
  }

  show_query()

  return PLUGIN_CONTINUE
}

abort_query()
{
  new playerId = queryPlayer[curQuery]
  end_query()
  startMenu(playerId)
}

end_query()
{
  recvattempts = 0
  multipart = 0
  query_in_progress = 0
  pendingQuery[queryPlayer[curQuery]] = 0
  curQuery = (curQuery+1)%MAX_QUERIES
  if (curQuery != lastQuery)
    set_task(0.1, "startQuery")
}

show_query()
{
  parse_rules()
  serverLastQuery[queryServer[curQuery]] = get_systime()
  showServer(queryPlayer[curQuery], queryServer[curQuery])

  end_query()
}

parse_rules()
{
  if (!equal(packet_buffer, {-1,-1,-1,-1,'E'}, 5))
  {
    log_amx("Invalid packet received to rule query.")
    client_print(queryPlayer[curQuery], print_chat, "Sorry an error occured while getting data from server.")
    return PLUGIN_CONTINUE
  }

  new rulenum = ((packet_buffer[6] & 0xff) << 8) | (packet_buffer[5] & 0xff)
  new packeti = 7
  new currule, strnum
  new foundtimeleft = 0
  for (currule = 0; currule < rulenum; currule++)
  {
    if (equal(packet_buffer[packeti], "amx_timeleft", 12))
    {
      packeti += 13
      copy(serverTimeleft[queryServer[curQuery]], MAX_TIME, packet_buffer[packeti])
      foundtimeleft = 1
      break
    }
    for (strnum=0; strnum < 2; strnum++)
    {
      while((packet_buffer[packeti] != 0) && (packeti < BUF_SIZE)) packeti++;
      packeti++;
    }
    if (packeti > BUF_SIZE)
    {
      break
    }
  }

  if (!foundtimeleft)
  {
    serverTimeleft[queryServer[curQuery]][0] = 0
  }

  return PLUGIN_CONTINUE
}

#define SERVER_INFO "\yServer info:^n\wName: %s^nPlaying %s with %d of %d players.^n"
#define SERVER_TIME "With %s minutes remaining.^n"

showServer(id, serveri)
{
  new menuString[500]
  new menuStrLen, menuKeys

  menuStrLen = format(menuString, 500, SERVER_INFO, serverName[serveri], serverMapName[serveri], serverPlayers[serveri], serverMaxPlayers[serveri])
  if (serverTimeleft[serveri][0])
  {
    serverTimeleft[serveri][2] = 0
    menuStrLen += format(menuString[menuStrLen], (500-menuStrLen), SERVER_TIME, serverTimeleft[serveri])
  }
  if (serverPlayers[serveri] < serverMaxPlayers[serveri])
  {
    format(menuString[menuStrLen], (500-menuStrLen), "^n1. Connect to server^n0. Back")
    menuKeys = (1<<0)|(1<<9)
  } else
  {
    format(menuString[menuStrLen], (500-menuStrLen), "^n0. Back")
    menuKeys = (1<<9)
  }

  lastServer[id] = serveri
  show_menu(id, menuKeys, menuString, -1)
  return PLUGIN_CONTINUE
}

public serverAction(id, key)
{
  switch(key)
  {
    case 0: 
    {
      new username[50], userauthid[35]
      get_user_name(id, username, 50)
      get_user_authid(id, userauthid, 35)
      log_amx("Sending %s<%s> to %s:%d", username, userauthid, serverHost[lastServer[id]], serverPort[lastServer[id]])
      if (serverPassword[lastServer[id]][0])
      {
        client_cmd(id, "password %s;wait;connect %s:%d", serverPassword[lastServer[id]], serverHost[lastServer[id]], serverPort[lastServer[id]])
      } else
      { 
        client_cmd(id, "connect %s:%d", serverHost[lastServer[id]], serverPort[lastServer[id]])
      }
    }
    case 9:
      showMenu(id)
  }

  return PLUGIN_HANDLED
}
